# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import numpy as np
from tqdm import tqdm_notebook
from MDAnalysis.analysis.distances import distance_array
from scipy.signal import savgol_filter
from mdgo.util import atom_vec
# from itertools import groupby

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


def trajectory(nvt_run, li_atom, run_start, run_end, species, selection_dict,
               distance):
    A_values = {}
    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    if species not in list(selection_dict):
        print('Invalid species selection')
        return None
    for ts in trj_analysis:
        selection = "(" + selection_dict[species] + ") and (around " \
                    + str(distance) + " index " \
                    + str(li_atom.id - 1) + ")"
        shell = nvt_run.select_atoms(selection, periodic=True)
        for atom in shell.atoms:
            if str(atom.id) not in A_values:
                A_values[str(atom.id)] = np.full(run_end-run_start, 100.)
        time_count += 1
    #print(A_values.keys())
    time_count = 0
    for ts in trj_analysis:
        for atomid in A_values.keys():
            dist = distance_array(ts[li_atom.id-1], ts[(int(atomid)-1)], ts.dimensions)
            A_values[atomid][time_count] = dist
        time_count += 1
    return A_values


def find_nearest(trj, time_step, distance, hopping_cutoff, smooth=51):
    time_span = len(list(trj.values())[0])
    for kw in list(trj):
        trj[kw] = savgol_filter(trj.get(kw), smooth, 2)
    site_distance = [100 for _ in range(time_span)]
    sites = [0 for _ in range(time_span)]
    sites[0] = min(trj, key=lambda k: trj[k][0])
    site_distance[0] = trj.get(sites[0])[0]
    for time in range(1, time_span):
        if sites[time - 1] == 0:
            old_site_distance = 100
        else:
            old_site_distance = trj.get(sites[time - 1])[time]
        if old_site_distance > hopping_cutoff:
            new_site = min(trj, key=lambda k: trj[k][time])
            new_site_distance = trj.get(new_site)[time]
            if new_site_distance > distance:
                site_distance[time] = 100
            else:
                sites[time] = new_site
                site_distance[time] = new_site_distance
        else:
            sites[time] = sites[time-1]
            site_distance[time] = old_site_distance
    sites = [int(i) for i in sites]
    sites_and_distance_array = np.array([[sites[i], site_distance[i]]
                                         for i in range(len(sites))])
    steps = []
    closest_step = 0
    previous_site = sites_and_distance_array[0][0]
    for i, step in enumerate(sites_and_distance_array):
        site = step[0]
        distance = step[1]
        if site == 0:
            pass
        else:
            if site == previous_site:
                if distance < sites_and_distance_array[closest_step][1]:
                    closest_step = i
                else:
                    pass
            else:
                steps.append(closest_step)
                closest_step = i
                previous_site = site
    if previous_site is not None:
        steps.append(closest_step)
    # change = len(steps) - 1
    change = (np.diff([i for i in sites if i != 0]) != 0).sum()
    frequency = change/(time_span * time_step)
    return sites, frequency, steps


def heat_map(nvt_run, li_atom, sites, dist_to_center, bind_atom_type,
             cartesian_by_ref, run_start, run_end):
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    coordinates = []
    for i, ts in enumerate(trj_analysis):
        if sites[i] == 0:
            pass
        else:
            center_atom = nvt_run.select_atoms("index " + str(sites[i] - 1))[0]
            selection = "(" + bind_atom_type + ") and " + "(around " \
                        + str(dist_to_center) + " index " \
                        + str(center_atom.id - 1) + ")"
            bind_atoms = nvt_run.select_atoms(selection, periodic=True)
            distances = distance_array(ts[li_atom.id - 1], bind_atoms.positions,
                                       ts.dimensions)
            idx = np.argpartition(distances[0], 3)
            vertex_atoms = bind_atoms[idx[:3]]
            vector_li = atom_vec(li_atom, center_atom, ts.dimensions)
            vector_a = atom_vec(vertex_atoms[0], center_atom, ts.dimensions)
            vector_b = atom_vec(vertex_atoms[1], center_atom, ts.dimensions)
            vector_c = atom_vec(vertex_atoms[2], center_atom, ts.dimensions)
            basis_abc = np.transpose([vector_a, vector_b, vector_c])
            abc_li = np.linalg.solve(basis_abc, vector_li)
            unit_x = np.linalg.norm(cartesian_by_ref[0, 0] * vector_a
                                    + cartesian_by_ref[0, 1] * vector_b
                                    + cartesian_by_ref[0, 2] * vector_c)
            unit_y = np.linalg.norm(cartesian_by_ref[1, 0] * vector_a
                                    + cartesian_by_ref[1, 1] * vector_b
                                    + cartesian_by_ref[1, 2] * vector_c)
            unit_z = np.linalg.norm(cartesian_by_ref[2, 0] * vector_a
                                    + cartesian_by_ref[2, 1] * vector_b
                                    + cartesian_by_ref[2, 2] * vector_c)
            vector_x = cartesian_by_ref[0] / unit_x
            vector_y = cartesian_by_ref[1] / unit_y
            vector_z = cartesian_by_ref[2] / unit_z
            basis_xyz = np.transpose([vector_x, vector_y, vector_z])
            xyz_li = np.linalg.solve(basis_xyz, abc_li)
            coordinates.append(xyz_li)
    return np.array(coordinates)


def get_full_coords(coords, reflection=None, rotation=None, inversion=None,
                    sample=None):
    coords_full = coords
    if reflection:
        for vec in reflection:
            coords_full = np.concatenate((coords, coords * vec), axis=0)
    if rotation:
        coords_copy = coords_full
        for mat in rotation:
            coords_rot = np.dot(coords_copy, mat)
            coords_full = np.concatenate((coords_full, coords_rot), axis=0)
    if inversion:
        coords_copy = coords_full
        for mat in inversion:
            coords_inv = np.dot(coords_copy, mat)
            coords_full = np.concatenate((coords_full, coords_inv), axis=0)
    if sample:
        index = np.random.choice(coords_full.shape[0], sample, replace=False)
        coords_full = coords_full[index]
    return coords_full


def cluster_coordinates(nvt_run, select_dict, run_start, run_end, species,
                        distance, basis_vectors=None, cluster_center="center"):
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    cluster_center = nvt_run.select_atoms(select_dict[cluster_center],
                                          periodic=True)[0]
    selection = "(" + " or ".join([s for s in species]) \
                + ") and (around " + str(distance) + " index "\
                    + str(cluster_center.id - 1) + ")"
    print(selection)
    shell = nvt_run.select_atoms(selection, periodic=True)
    cluster = []
    for atom in shell:
        coord_list = []
        for ts in trj_analysis:
            coord_list.append(atom.position)
        cluster.append(np.mean(np.array(coord_list), axis=0))
    cluster = np.array(cluster)
    if basis_vectors:
        if len(basis_vectors) == 2:
            vec1 = basis_vectors[0]
            vec2 = basis_vectors[1]
            vec3 = np.cross(vec1, vec2)
            vec2 = np.cross(vec1, vec3)
        elif len(basis_vectors) == 3:
            vec1 = basis_vectors[0]
            vec2 = basis_vectors[1]
            vec3 = basis_vectors[2]
        else:
            raise ValueError("incorrect vector format")
        vec1 = vec1 / np.linalg.norm(vec1)
        vec2 = vec2 / np.linalg.norm(vec2)
        vec3 = vec3 / np.linalg.norm(vec3)
        basis_xyz = np.transpose([vec1, vec2, vec3])
        cluster_norm = np.linalg.solve(basis_xyz, cluster.T).T
        cluster_norm = cluster_norm - np.mean(cluster_norm, axis=0)
        return cluster_norm
    else:
        return cluster


def num_of_neighbor_one_li(nvt_run, li_atom, species, select_dict,
                           distance, run_start, run_end):

    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    if species in select_dict.keys():
        cn_values = np.zeros(int(len(trj_analysis)))
    else:
        print('Invalid species selection')
        return None
    for ts in trj_analysis:
        selection = "(" + select_dict[species] + ") and (around "\
                    + str(distance) + " index "\
                    + str(li_atom.id - 1) + ")"
        shell = nvt_run.select_atoms(selection, periodic=True)
        for _ in shell.atoms:
            cn_values[time_count] += 1
        time_count += 1
    return cn_values


def num_of_neighbor_one_li_multi(nvt_run, li_atom, species_list, select_dict,
                                 distances, run_start, run_end, write=False,
                                 structure_code=None, write_freq=0,
                                 write_path=None, element_id_dict=None):

    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    cn_values = dict()
    for kw in species_list:
        if kw in select_dict.keys():
            cn_values[kw] = np.zeros(int(len(trj_analysis)))
        else:
            print('Invalid species selection')
            return None
    cn_values["total"] = np.zeros(int(len(trj_analysis)))
    for ts in trj_analysis:
        digit_of_species = len(species_list) - 1
        for kw in species_list:
            selection = "(" + select_dict[kw] + ") and (around "\
                        + str(distances[kw]) + " index "\
                        + str(li_atom.id - 1) + ")"
            shell = nvt_run.select_atoms(selection, periodic=True)
            # for each atom in shell, create/add to dictionary
            # (key = atom id, value = list of values for step function)
            for _ in shell.atoms:
                cn_values[kw][time_count] += 1
                cn_values["total"][time_count] += 10**digit_of_species
            digit_of_species = digit_of_species - 1
        if write and cn_values["total"][time_count] == structure_code:
            a = np.random.random()
            if a > 1 - write_freq:
                print("writing")
                species = ' or '.join("(same resid as (" + select_dict[kw]
                                      + " and around " + str(distances[kw])
                                      + " index " + str(li_atom.id - 1)
                                      + "))" for kw in species_list)
                species = "((" + species + ")and not " \
                          + select_dict["cation"] + ")"
                structure = nvt_run.select_atoms(species,
                                                 periodic=True)
                li_pos = ts[(int(li_atom.id)-1)]
                path = write_path + str(li_atom.id) + "_" + str(int(ts.time)) \
                    + "_" + str(structure_code) + ".xyz"
                write_out(li_pos, structure, element_id_dict, path)
        time_count += 1
    return cn_values


def num_of_neighbor_one_li_simple(nvt_run, li_atom, species, select_dict,
                                  distance, run_start, run_end):

    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    if species in select_dict.keys():
        cn_values = np.zeros(int(len(trj_analysis)))
    else:
        print('Invalid species selection')
        return None
    for ts in trj_analysis:
        selection = "(" + select_dict[species] + ") and (around "\
                    + str(distance) + " index "\
                    + str(li_atom.id - 1) + ")"
        shell = nvt_run.select_atoms(selection, periodic=True)
        shell_len = len(shell)
        if shell_len == 0:
            cn_values[time_count] = 1
        elif shell_len == 1:
            selection_species = "(" + select_dict["cation"] + " and around " + \
                                str(distance) + " index " + \
                                str(shell.atoms[0].id - 1) + ")"
            shell_species = nvt_run.select_atoms(selection_species,
                                                 periodic=True)
            shell_species_len = len(shell_species) - 1
            if shell_species_len == 0:
                cn_values[time_count] = 2
            else:
                cn_values[time_count] = 3
        else:
            cn_values[time_count] = 3
        time_count += 1
    return cn_values


def num_of_neighbor_one_li_simple_extra(nvt_run, li_atom, species, select_dict,
                                        distance, run_start, run_end):

    time_count = 0
    emc_angle = list()
    ec_angle = list()
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    if species in select_dict.keys():
        cn_values = np.zeros(int(len(trj_analysis)))
    else:
        print('Invalid species selection')
        return None
    for ts in trj_analysis:
        selection = "(" + select_dict[species] + ") and (around "\
                    + str(distance) + " index "\
                    + str(li_atom.id - 1) + ")"
        shell = nvt_run.select_atoms(selection, periodic=True)
        shell_len = len(shell)
        if shell_len == 0:
            cn_values[time_count] = 1
        elif shell_len == 1:
            selection_species = "(" + select_dict["cation"] + " and around " + \
                                str(distance) + " index " + \
                                str(shell.atoms[0].id - 1) + ")"
            shell_species = nvt_run.select_atoms(selection_species,
                                                 periodic=True)
            shell_species_len = len(shell_species) - 1
            if shell_species_len == 0:
                cn_values[time_count] = 2
                li_pos = li_atom.position
                p_pos = shell.atoms[0].position
                ec_select = "(" + select_dict["EC"] + ") and (around "\
                    + str(3) + " index "\
                    + str(li_atom.id - 1) + ")"
                emc_select = "(" + select_dict["EMC"] + ") and (around "\
                    + str(3) + " index "\
                    + str(li_atom.id - 1) + ")"
                ec_group = nvt_run.select_atoms(ec_select, periodic=True)
                emc_group = nvt_run.select_atoms(emc_select, periodic=True)
                for atom in ec_group.atoms:
                    theta = angle(p_pos, li_pos, atom.position)
                    ec_angle.append(theta)
                for atom in emc_group.atoms:
                    theta = angle(p_pos, li_pos, atom.position)
                    emc_angle.append(theta)
            else:
                cn_values[time_count] = 3
        else:
            cn_values[time_count] = 3
        time_count += 1
    return cn_values, np.array(ec_angle), np.array(emc_angle)


def num_of_neighbor_one_li_simple_extra_two(nvt_run, li_atom, species_list,
                                            select_dict, distances, run_start,
                                            run_end):
    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    cip_step = list()
    ssip_step = list()
    agg_step = list()
    cn_values = dict()
    for kw in species_list:
        if kw in select_dict.keys():
            cn_values[kw] = np.zeros(int(len(trj_analysis)))
        else:
            print('Invalid species selection')
            return None
    cn_values["total"] = np.zeros(int(len(trj_analysis)))
    for ts in trj_analysis:
        digit_of_species = len(species_list) - 1
        for kw in species_list:
            selection = "(" + select_dict[kw] + ") and (around "\
                        + str(distances[kw]) + " index "\
                        + str(li_atom.id - 1) + ")"
            shell = nvt_run.select_atoms(selection, periodic=True)
            # for each atom in shell, create/add to dictionary
            # (key = atom id, value = list of values for step function)
            for _ in shell.atoms:
                cn_values[kw][time_count] += 1
                cn_values["total"][time_count] += 10**digit_of_species
            digit_of_species = digit_of_species - 1

        selection = "(" + select_dict["anion"] + ") and (around "\
                    + str(distances["anion"]) + " index "\
                    + str(li_atom.id - 1) + ")"
        shell = nvt_run.select_atoms(selection, periodic=True)
        shell_len = len(shell)
        if shell_len == 0:
            ssip_step.append(time_count)
        elif shell_len == 1:
            selection_species = "(" + select_dict["cation"] + " and around " + \
                                str(distances["anion"]) + " index " + \
                                str(shell.atoms[0].id - 1) + ")"
            shell_species = nvt_run.select_atoms(selection_species,
                                                 periodic=True)
            shell_species_len = len(shell_species) - 1
            if shell_species_len == 0:
                cip_step.append(time_count)
            else:
                agg_step.append(time_count)
        else:
            agg_step.append(time_count)
        time_count += 1
    cn_ssip = dict()
    cn_cip = dict()
    cn_agg = dict()
    for kw in species_list:
        cn_ssip[kw] = np.mean(cn_values[kw][ssip_step])
        cn_cip[kw] = np.mean(cn_values[kw][cip_step])
        cn_agg[kw] = np.mean(cn_values[kw][agg_step])
    return cn_ssip, cn_cip, cn_agg


def angle(a, b, c):
    ba = a - b
    bc = c - b
    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
    cosine_angle = np.clip(cosine_angle, -1.0, 1.0)
    angle = np.arccos(cosine_angle)
    return np.degrees(angle)


# Depth-first traversal
def num_of_neighbor_one_li_complex(nvt_run, li_atom, species, selection_dict,
                                   distance, run_start, run_end):
    time_count = 0
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    A_values = np.zeros((int(len(trj_analysis)), 4))
    for ts in trj_analysis:
        li_list = [li_atom.id]
        p_list = []
        shell = nvt_run.select_atoms("(" + selection_dict[species] +
                                     " and around " + str(distance)
                                     + " index " + str(li_atom.id - 1) + ")",
                                     periodic=True)
        for atom in shell.atoms:
            if atom.resid not in p_list:
                p_list.append(atom.resid)
                A_values[time_count][0] += 1
                shell_pf6 = nvt_run.select_atoms("(type 17 and around 3 resid "
                                                 + str(atom.resid) + ")",
                                                 periodic=True)
                for li in shell_pf6:
                    if li.id not in li_list:
                        li_list.append(li.id)
                        A_values[time_count][1] += 1
                        shell_li = nvt_run.select_atoms("(type 15 and around 3 index " + str(li.id - 1) + ")",
                                                        periodic=True)
                        for p in shell_li.atoms:
                            if p.resid not in p_list:
                                p_list.append(p.resid)
                                A_values[time_count][2] += 1
                                shell_p = nvt_run.select_atoms("(type 17 and around 3 resid " + str(p.resid) + ")",
                                                               periodic=True)
                                for moreli in shell_pf6:
                                    if moreli.id not in li_list:
                                        li_list.append(moreli.id)
                                        A_values[time_count][3] += 1


def coord_shell_array(nvt_run, func, li_atoms, species, select_dict,
                      distance, run_start, run_end):
    """
    Args:
        nvt_run: MDAnalysis Universe
        func: One of the neighbor statistical method
        li_atoms: atom group of the Li atoms.
    """
    num_array = func(nvt_run, li_atoms[0], species, select_dict,
                     distance, run_start, run_end)
    if isinstance(num_array, np.ndarray):
        for li in tqdm_notebook(li_atoms[1::]):
            this_li = func(nvt_run, li, species, select_dict,
                           distance, run_start, run_end)
            num_array = np.concatenate((num_array, this_li), axis=0)
    else:
        for li in tqdm_notebook(li_atoms[1::]):
            this_li = func(nvt_run, li, species, select_dict,
                           distance, run_start, run_end)
            for kw in num_array.keys():
                num_array[kw] = np.concatenate((num_array.get(kw),
                                                this_li.get(kw)), axis=0)
    return num_array


def write_out(li_pos, selection, element_id_dict, path):
    lines = list()
    lines.append(str(len(selection) + 1))
    lines.append("")
    lines.append("Li 0.0000000 0.0000000 0.0000000")
    box = selection.dimensions
    half_box = np.array([box[0], box[1], box[2]])/2
    for atom in selection:
        locs = list()
        for i in range(3):
            loc = atom.position[i] - li_pos[i]
            if loc > half_box[i]:
                loc = loc - box[i]
            elif loc < -half_box[i]:
                loc = loc + box[i]
            else:
                pass
            locs.append(loc)
        line = element_id_dict.get(int(atom.type)) \
            + ' ' + ' '.join(str(loc) for loc in locs)
        lines.append(line)
    with open(path, "w") as xyz_file:
        xyz_file.write("\n".join(lines))



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import numpy as np
from io import StringIO
import re
import pandas as pd
import math

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"

MM_of_Elements = {'H': 1.00794, 'He': 4.002602, 'Li': 6.941, 'Be': 9.012182,
                  'B': 10.811, 'C': 12.0107, 'N': 14.0067, 'O': 15.9994,
                  'F': 18.9984032, 'Ne': 20.1797, 'Na': 22.98976928,
                  'Mg': 24.305, 'Al': 26.9815386, 'Si': 28.0855, 'P': 30.973762,
                  'S': 32.065, 'Cl': 35.453, 'Ar': 39.948, 'K': 39.0983,
                  'Ca': 40.078, 'Sc': 44.955912, 'Ti': 47.867, 'V': 50.9415,
                  'Cr': 51.9961, 'Mn': 54.938045, 'Fe': 55.845, 'Co': 58.933195,
                  'Ni': 58.6934, 'Cu': 63.546, 'Zn': 65.409, 'Ga': 69.723,
                  'Ge': 72.64, 'As': 74.9216, 'Se': 78.96, 'Br': 79.904,
                  'Kr': 83.798, 'Rb': 85.4678, 'Sr': 87.62, 'Y': 88.90585,
                  'Zr': 91.224, 'Nb': 92.90638, 'Mo': 95.94, 'Tc': 98.9063,
                  'Ru': 101.07, 'Rh': 102.9055, 'Pd': 106.42, 'Ag': 107.8682,
                  'Cd': 112.411, 'In': 114.818, 'Sn': 118.71, 'Sb': 121.760,
                  'Te': 127.6, 'I': 126.90447, 'Xe': 131.293, 'Cs': 132.9054519,
                  'Ba': 137.327, 'La': 138.90547, 'Ce': 140.116,
                  'Pr': 140.90465, 'Nd': 144.242, 'Pm': 146.9151, 'Sm': 150.36,
                  'Eu': 151.964, 'Gd': 157.25, 'Tb': 158.92535, 'Dy': 162.5,
                  'Ho': 164.93032, 'Er': 167.259, 'Tm': 168.93421, 'Yb': 173.04,
                  'Lu': 174.967, 'Hf': 178.49, 'Ta': 180.9479, 'W': 183.84,
                  'Re': 186.207, 'Os': 190.23, 'Ir': 192.217, 'Pt': 195.084,
                  'Au': 196.966569, 'Hg': 200.59, 'Tl': 204.3833, 'Pb': 207.2,
                  'Bi': 208.9804, 'Po': 208.9824, 'At': 209.9871,
                  'Rn': 222.0176, 'Fr': 223.0197, 'Ra': 226.0254,
                  'Ac': 227.0278, 'Th': 232.03806, 'Pa': 231.03588,
                  'U': 238.02891, 'Np': 237.0482, 'Pu': 244.0642,
                  'Am': 243.0614, 'Cm': 247.0703, 'Bk': 247.0703,
                  'Cf': 251.0796, 'Es': 252.0829, 'Fm': 257.0951,
                  'Md': 258.0951, 'No': 259.1009, 'Lr': 262, 'Rf': 267,
                  'Db': 268, 'Sg': 271, 'Bh': 270, 'Hs': 269, 'Mt': 278,
                  'Ds': 281, 'Rg': 281, 'Cn': 285, 'Nh': 284, 'Fl': 289,
                  'Mc': 289, 'Lv': 292, 'Ts': 294, 'Og': 294, 'ZERO': 0}

SECTION_SORTER = {
    "atoms": {
        "in_kw": None,
        "in_header": ["atom", "charge", "sigma", "epsilon"],
        "sec_number": None,
        "desired_split": None,
        "desired_cols": None,
        "out_kw": None,
        "ff_header": ["epsilon", "sigma"],
        "topo_header": ["mol-id", "type", "charge", "x", "y", "z"]
    },
    "bonds": {
        "in_kw": "Stretch",
        "in_header": ["atom1", "atom2", "k", "r0"],
        "sec_number": 5,
        "desired_split": 2,
        "desired_cols": 4,
        "out_kw": ["Bond Coeffs", "Bonds"],
        "ff_header": ["k", "r0"],
        "topo_header": ["type", "atom1", "atom2"]
    },
    "angles": {
        "in_kw": "Bending",
        "in_header": ["atom1", "atom2", "atom3", "k", "theta0"],
        "sec_number": 6,
        "desired_split": 1,
        "desired_cols": 5,
        "out_kw": ["Angle Coeffs", "Angles"],
        "ff_header": ["k", "theta0"],
        "topo_header": ["type", "atom1", "atom2", "atom3"]
    },
    "dihedrals": {
        "in_kw": "proper Torsion",
        "in_header": [
            "atom1", "atom2", "atom3", "atom4", "v1", "v2", "v3", "v4"
        ],
        "sec_number": 7,
        "desired_split": 1,
        "desired_cols": 8,
        "out_kw": ["Dihedral Coeffs", "Dihedrals"],
        "ff_header": ["v1", "v2", "v3", "v4"],
        "topo_header": ["type", "atom1", "atom2", "atom3", "atom4"]
    },
    "impropers": {
        "in_kw": "improper Torsion",
        "in_header": ["atom1", "atom2", "atom3", "atom4", "v2"],
        "sec_number": 8,
        "desired_split": 1,
        "desired_cols": 5,
        "out_kw": ["Improper Coeffs", "Impropers"],
        "ff_header": ["v1", "v2", "v3", "v4"],
        "topo_header": ["type", "atom1", "atom2", "atom3", "atom4"]
    },
}

BOX = """{0:6f} {1:6f} xlo xhi
{0:6f} {1:6f} ylo yhi
{0:6f} {1:6f} zlo zhi"""


def atom_vec(atom1, atom2, dimension):
    """
    Calculate the vector of the positions from atom2 to atom1.
    """
    vec = [0, 0, 0]
    for i in range(3):
        diff = atom1.position[i]-atom2.position[i]
        if diff > dimension[i]/2:
            vec[i] = diff - dimension[i]
        elif diff < - dimension[i]/2:
            vec[i] = diff + dimension[i]
        else:
            vec[i] = diff
    return np.array(vec)


def position_vec(pos1, pos2, dimension):
    """
    Calculate the vector from pos2 to pos2.
    """
    vec = [0, 0, 0]
    for i in range(3):
        diff = pos1[i]-pos2[i]
        if diff > dimension[i]/2:
            vec[i] = diff - dimension[i]
        elif diff < - dimension[i]/2:
            vec[i] = diff + dimension[i]
        else:
            vec[i] = diff
    return np.array(vec)


def mass_to_name(df):
    """
    Create a dict for mapping atom type id to element from the mass information.

    Args:
        df (pandas.DataFrame): The masses attribute from LammpsData object
    Return:
        dict: The element dict.
    """
    atoms = {}
    for row in df.index:
        for item in MM_of_Elements.items():
            if math.isclose(df["mass"][row], item[1], abs_tol=0.01):
                atoms[row] = item[0]
    return atoms


def ff_parser(ff_dir, xyz_dir):
    """
    A parser to convert a force field field from Maestro format
    to LAMMPS data format.

    Args:
        ff_dir (str): The path to the Maestro force field file.
        xyz_dir (str): The path to the xyz structure file.
    Return:
        str: The output LAMMPS data string.
    """
    with open(xyz_dir, 'r') as f_xyz:
        molecule = pd.read_table(
            f_xyz,
            skiprows=2,
            delim_whitespace=True,
            names=['atom', 'x', 'y', 'z']
        )
        coordinates = molecule[["x", "y", "z"]]
        lo = coordinates.min().min() - 0.5
        hi = coordinates.max().max() + 0.5
    with open(ff_dir, 'r') as f:
        lines_org = f.read()
        lines = lines_org.split("\n\n")
        atoms = "\n".join(lines[4].split("\n", 4)[4].split("\n")[:-1])
        dfs = dict()
        dfs["atoms"] = pd.read_csv(
            StringIO(atoms),
            names=SECTION_SORTER.get("atoms").get("in_header"),
            delim_whitespace=True,
            usecols=[0, 4, 5, 6]
        )
        dfs["atoms"] = pd.concat(
            [dfs["atoms"], coordinates],
            axis=1
        )
        dfs["atoms"].index += 1
        dfs["atoms"].index.name = "type"
        dfs["atoms"] = dfs["atoms"].reset_index()
        dfs["atoms"].index += 1
        types = dfs["atoms"].copy().reset_index().set_index('atom')['type']
        replace_dict = {
            "atom1": dict(types),
            "atom2": dict(types),
            "atom3": dict(types),
            "atom4": dict(types)
        }
        counts = dict()
        counts["atoms"] = len(dfs["atoms"].index)
        mass_list = list()
        for index, row in dfs["atoms"].iterrows():
            mass_list.append(
                MM_of_Elements.get(re.split(r'(\d+)', row['atom'])[0])
            )
        mass_df = pd.DataFrame(mass_list)
        mass_df.index += 1
        mass_string = mass_df.to_string(
            header=False,
            index_names=False,
            float_format="{:.3f}".format
        )
        masses = ["Masses", mass_string]
        ff = ["Pair Coeffs"]
        dfs["atoms"]["mol-id"] = 1
        atom_ff_string = dfs["atoms"][
            SECTION_SORTER["atoms"]["ff_header"]
        ].to_string(
            header=False,
            index_names=False
        )
        ff.append(atom_ff_string)
        topo = ["Atoms"]
        atom_topo_string = dfs["atoms"][
            SECTION_SORTER["atoms"]["topo_header"]
        ].to_string(
            header=False,
            index_names=False
        )
        topo.append(atom_topo_string)
        for section in list(SECTION_SORTER.keys())[1:]:
            if SECTION_SORTER[section]["in_kw"] in lines_org:
                a, b, c, d = SECTION_SORTER[section]["sec_number"],\
                             SECTION_SORTER[section]["desired_split"],\
                             SECTION_SORTER[section]["desired_cols"],\
                             SECTION_SORTER[section]["in_header"]
                section_str = lines[a].split("\n", b)[b]
                dfs[section] = pd.read_csv(
                    StringIO(section_str),
                    names=d,
                    delim_whitespace=True,
                    usecols=list(range(c)),
                )

                dfs[section].index += 1
                dfs[section].index.name = "type"
                dfs[section] = dfs[section].replace(replace_dict)
                dfs[section] = dfs[section].reset_index()
                dfs[section].index += 1
                if section == "impropers":
                    dfs[section]["v1"] = 0.0
                    dfs[section]["v3"] = 0.0
                    dfs[section]["v4"] = 0.0
                ff_string = dfs[section][
                    SECTION_SORTER[section]["ff_header"]
                ].to_string(
                    header=False,
                    index_names=False
                )
                ff.append(SECTION_SORTER[section]["out_kw"][0])
                ff.append(ff_string)
                topo_string = dfs[section][
                    SECTION_SORTER[section]["topo_header"]
                ].to_string(
                    header=False,
                    index_names=False
                )
                topo.append(SECTION_SORTER[section]["out_kw"][1])
                topo.append(topo_string)
                counts[section] = len(dfs[section].index)
        max_stats = len(str(max(list(counts.values()))))
        stats_template = "{:>%d}  {}" % max_stats
        count_lines = [stats_template.format(v, k) for k, v in counts.items()]
        type_lines = [
            stats_template.format(v, k[:-1] + " types")
            for k, v in counts.items()
        ]
        stats = "\n".join(count_lines + [""] + type_lines)
        header = [
            f"LAMMPS data file created by mdgo (by {__author__})\n"
            "# OPLS force field: harmonic, harmonic, opls, opls",
            stats,
            BOX.format(lo, hi)
        ]
        data_string = "\n\n".join(header + masses + ff + topo) + "\n"
        return data_string



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

try:
    import MDAnalysis.analysis.msd as msd
except ImportError:
    msd = None


import numpy as np
from tqdm import tqdm_notebook

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


def total_msd(nvt_run, start, stop, select='all', msd_type='xyz', fft=True):
    if msd is not None:
        msd_calculator = msd.EinsteinMSD(nvt_run, select=select,
                                         msd_type=msd_type, fft=fft)
        msd_calculator.run(start=start, stop=stop)
        total_array = msd_calculator.timeseries
        return total_array
    else:
        if fft:
            print(
                "Warning! MDAnalysis version too low, fft not supported. "
                "Use conventional instead"
            )
        return _total_msd(nvt_run, select, start, stop)


def _total_msd(nvt_run, select, run_start, run_end):
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    li_atoms = nvt_run.select_atoms(select)
    all_list = list()
    for li_atom in li_atoms:
        coords = list()
        for ts in trj_analysis:
            current_coord = ts[li_atom.id - 1]
            coords.append(current_coord)
        all_list.append(np.array(coords))
    total_array = msd_states(all_list, run_end - run_start - 1)
    return total_array


def msd_states(coord_list, largest):
    msd_dict = dict()
    for state in coord_list:
        n_frames = state.shape[0]
        lag_times = np.arange(1, min(n_frames, largest))
        for lag in lag_times:
            disp = state[:-lag, :] - state[lag:, :]
            sqdist = np.square(disp).sum(axis=-1)
            if lag in msd_dict.keys():
                msd_dict[lag] = np.concatenate((msd_dict[lag], sqdist), axis=0)
            else:
                msd_dict[lag] = sqdist
    timeseries = list()
    time_range = len(msd_dict) + 1
    msds_by_state = np.zeros(time_range)
    for kw in range(1, time_range):
        msds_by_state[kw] = msd_dict.get(kw).mean()
        timeseries.append(msds_by_state[kw])
    timeseries = np.array(timeseries)
    return timeseries


def states_coord_array(nvt_run, li_atom, select_dict, distance,
                       run_start, run_end):
    trj_analysis = nvt_run.trajectory[run_start:run_end:]
    attach_list = list()
    free_list = list()
    coords = list()
    prev_state = None
    prev_coord = None
    for ts in trj_analysis:
        selection = "(" + select_dict["anion"] + ") and (around " \
                    + str(distance) + " index " \
                    + str(li_atom.id - 1) + ")"
        shell_anion = nvt_run.select_atoms(selection, periodic=True)
        current_state = 0
        if len(shell_anion) > 0:
            current_state = 1
        current_coord = ts[li_atom.id-1]

        if prev_state:
            if current_state == prev_state:
                coords.append(current_coord)
                prev_coord = current_coord
            else:
                if len(coords) > 1:
                    if prev_state:
                        attach_list.append(np.array(coords))
                    else:
                        free_list.append(np.array(coords))
                prev_state = current_state
                coords = list()
                coords.append(prev_coord)
                coords.append(current_coord)
                prev_coord = current_coord
        else:
            coords.append(current_coord)
            prev_state = current_state
            prev_coord = current_coord

    if len(coords) > 1:
        if prev_state:
            attach_list.append(np.array(coords))
        else:
            free_list.append(np.array(coords))

    return attach_list, free_list


def partial_msd(nvt_run, li_atoms, largest, select_dict, distance,
                run_start, run_end):
    free_coords = list()
    attach_coords = list()
    for i in tqdm_notebook(list(range(len(li_atoms)))):
        attach_coord, free_coord = states_coord_array(nvt_run, li_atoms[i],
                                                      select_dict, distance,
                                                      run_start, run_end)
        attach_coords.extend(attach_coord)
        free_coords.extend(free_coord)
    attach_data = None
    free_data = None
    if len(attach_coords) > 0:
        attach_data = msd_states(attach_coords, largest)
    if len(free_coords) > 0:
        free_data = msd_states(free_coords, largest)
    return free_data, attach_data


def msd_by_length(coord_list):
    msd_dict = dict()
    for state in coord_list:
        n_frames = state.shape[0]
        disp = state[0, :] - state[-1, :]
        sqdist = np.square(disp).sum()
        if n_frames in list(msd_dict):
            msd = msd_dict.get(n_frames)
            msd_dict[n_frames].append(sqdist)
        else:
            msd_dict[n_frames] = [sqdist]
    timespan = list(msd_dict)
    timespan.sort()
    timeseries = list()
    for time in timespan:
        timeseries.append(np.array([time, np.mean(msd_dict.get(time))]))
    timeseries = np.array(timeseries)
    return timeseries


def special_msd(nvt_run, li_atoms, select_dict, distance, run_start, run_end):
    free_coords = list()
    attach_coords = list()
    for i in tqdm_notebook(list(range(len(li_atoms)))):
        attach_coord, free_coord = states_coord_array(nvt_run, li_atoms[i],
                                                      select_dict, distance,
                                                      run_start, run_end)
        attach_coords.extend(attach_coord)
        free_coords.extend(free_coord)
    attach_data = None
    free_data = None
    if len(attach_coords) > 0:
        attach_data = msd_by_length(attach_coords)
    if len(free_coords) > 0:
        free_data = msd_by_length(free_coords)
    return free_data, attach_data



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import MDAnalysis
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pymatgen.io.lammps.data import LammpsData
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.lib.distances import capped_distance
from tqdm import tqdm_notebook
from mdgo.util import mass_to_name
from mdgo.conductivity import calc_cond, conductivity_calculator
from mdgo.coordination import (
    coord_shell_array,
    num_of_neighbor_one_li,
    num_of_neighbor_one_li_multi,
    num_of_neighbor_one_li_simple,
    trajectory,
    find_nearest,
    heat_map,
    get_full_coords
)
from mdgo.msd import total_msd, partial_msd, special_msd
from mdgo.residence_time import calc_neigh_corr, fit_residence_time

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


class MdRun:
    """
    A core class for MD results analysis.

    Args:
        data_dir (str): Path to the data file.
        wrapped_dir (str): Path to the wrapped dcd file.
        unwrapped_dir (str): Path to the unwrapped dcd file.
        nvt_start (int): NVT start time step.
        time_step (int or float): LAMMPS timestep.
        name (str): Name of the MD run.
        select_dict: A dictionary of species selection.
        cation_charge: Charge of cation. Default to 1.
        anion_charge: Charge of anion. Default to 1.
        cond (bool): Whether to calculate conductivity MSD. Default to True.
    """

    def __init__(self, data_dir, wrapped_dir, unwrapped_dir, nvt_start,
                 time_step, name, select_dict, cation_charge=1, anion_charge=-1,
                 cond=True):
        """
        Base constructor.


        """

        self.wrapped_run = MDAnalysis.Universe(data_dir,
                                               wrapped_dir,
                                               format="LAMMPS")
        self.unwrapped_run = MDAnalysis.Universe(data_dir,
                                                 unwrapped_dir,
                                                 format="LAMMPS")
        self.nvt_start = nvt_start
        self.time_step = time_step
        self.name = name
        self.data = LammpsData.from_file(data_dir)
        self.element_id_dict = mass_to_name(self.data.masses)
        self.select_dict = select_dict
        self.nvt_steps = self.wrapped_run.trajectory.n_frames
        self.time_array = [i * self.time_step for i in range(self.nvt_steps)]
        self.cation_charge = cation_charge
        self.anion_charge = anion_charge
        self.num_li = \
            len(self.wrapped_run.select_atoms(self.select_dict["cation"]))
        if cond:
            self.cond_array = self.get_cond_array()
        else:
            self.cond_array = None
        self.init_x = self.get_init_dimension()[0]
        self.init_y = self.get_init_dimension()[1]
        self.init_z = self.get_init_dimension()[2]
        self.init_v = self.init_x * self.init_y * self.init_z
        self.nvt_x = self.get_nvt_dimension()[0]
        self.nvt_y = self.get_nvt_dimension()[1]
        self.nvt_z = self.get_nvt_dimension()[2]
        self.nvt_v = self.nvt_x * self.nvt_y * self.nvt_z
        gas_constant = 8.314
        temp = 298.15
        faraday_constant_2 = 96485 * 96485
        self.c = (self.num_li / (self.nvt_v * 1e-30)) / (6.022*1e23)
        self.d_to_sigma = self.c * faraday_constant_2 / (gas_constant * temp)

    def get_init_dimension(self):
        """
        Returns the initial box dimension.
        """
        return self.wrapped_run.dimensions

    def get_nvt_dimension(self):
        """
        Returns the equilibrium box dimension.
        """
        return self.wrapped_run.trajectory[-1].dimensions

    def get_cond_array(self):
        """ Calculates the conductivity "mean square displacement".

        Returns an array of MSD values for each time in the trajectory.
        """
        nvt_run = self.unwrapped_run
        cations = nvt_run.select_atoms(self.select_dict["cation"])
        anions = nvt_run.select_atoms(self.select_dict["anion"])
        cond_array = calc_cond(nvt_run, anions, cations, self.nvt_start,
                               self.cation_charge, self.anion_charge)
        return cond_array

    def plot_cond_array(self, start, end, *runs):
        """ Plots the conductivity MSD as a function of time

            Args:
                start (int): Start time step.
                end (int): End time step.
                runs (MdRun): Other runs to be compared in the same plot.
        """
        if self.cond_array is None:
            self.cond_array = self.get_cond_array()
        colors = ["g", "r", "c", "m", "y", "k"]
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.loglog(self.time_array[start:end],
                  self.cond_array[start:end], color="b", lw=2,
                  label=self.name)
        for i, run in enumerate(runs):
            ax.loglog(run.time_array[start:end],
                      run.cond_array[start:end], color=colors[i], lw=2,
                      label=run.name)
        ax.set_ylabel('MSD (A^2)')
        ax.set_xlabel('Time (ps)')
        ax.set_ylim([10, 1000000])
        ax.set_xlim([100, 500000000])
        ax.legend()
        fig.show()

    def get_conductivity(self, start, end):
        """ Calculates the Green-Kubo (GK) conductivity

        Args:
            start (int): Start time step.
            end (int): End time step.

        Print conductivity in mS/cm.
        """
        conductivity_calculator(self.time_array, self.cond_array,
                                self.nvt_v, self.name, start, end)
        return None

    def coord_num_array_one_species(self, species, distance,
                                    run_start, run_end):
        """ Calculates the coordination number array of one species around
        the cation.

        Args:
            species (str): The interested species.
            distance (int or float): The coordination cutoff distance.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns an array of the species coordination number for each time
        in the trajectory.
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        num_array = coord_shell_array(nvt_run, num_of_neighbor_one_li,
                                      li_atoms, species, self.select_dict,
                                      distance, run_start, run_end)
        return num_array

    def coord_num_array_multi_species(self, species, distances,
                                      run_start, run_end):
        """ Calculates the coordination number array of multiple species around
        the cation

        Args:
            species (str): The interested species.
            distances (dict): A dict of coordination cutoff distance of species.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns a python dict of arrays of coordination numbers of each species
         for each time in the trajectory.
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        num_array = coord_shell_array(nvt_run, num_of_neighbor_one_li_multi,
                                      li_atoms, species, self.select_dict,
                                      distances, run_start, run_end)
        return num_array

    def get_solvation_structure(self, species, distances, run_start, run_end,
                                structure_code, write_freq, write_path):
        """ Writes out the desired solvation structure

        Args:
            species (str): The interested species.
            distances (dict): A dict of coordination cutoff distance of species.
            run_start (int): Start time step.
            run_end (int): End time step.
            structure_code: An integer code representing the solvation
                structure, for example, 221 is two species A, two species B
                and one species C.
            write_freq: Probability to write out files.
            write_path: Path to write out files.

        Write out a series of solvation structures as .xyz files .
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        for li in tqdm_notebook(li_atoms):
            num_of_neighbor_one_li_multi(nvt_run, li, species, self.select_dict,
                                         distances, run_start, run_end,
                                         write=True,
                                         structure_code=structure_code,
                                         write_freq=write_freq,
                                         write_path=write_path,
                                         element_id_dict=self.element_id_dict)

    def coord_num_array_simple(self, species, distance, run_start, run_end):
        """ Calculates the solvation structure type (1 for ssIP, 2 for CIP,
        3 for AGG) array of the cation

        Args:
            species (str): The interested species.
            distance (int or float): The coordination cutoff distance.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns an array of the solvation structure type
        for each time in the trajectory.
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        num_array = coord_shell_array(nvt_run, num_of_neighbor_one_li_simple,
                                      li_atoms, species, self.select_dict,
                                      distance, run_start, run_end)
        return num_array

    def coordination_one_species(self, species, distance, run_start, run_end):
        """ Tabulates the coordination number distribution of one species
        around the cation

        Args:
            species (str): The interested species.
            distance (int or float): The coordination cutoff distance.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns pandas.dataframe of the species coordination number
        and corresponding percentage.
        """
        num_array = self.coord_num_array_one_species(species, distance,
                                                     run_start, run_end)
        shell_component, shell_count = np.unique(num_array.flatten(),
                                                 return_counts=True)
        combined = np.vstack((shell_component, shell_count)).T

        item_name = "Num of " + species + " within " + str(distance) + " " \
                    + "\u212B"
        item_list = []
        percent_list = []
        for i in range(len(combined)):
            item_list.append(str(int(combined[i, 0])))
            percent_list.append(str("%.4f" % (combined[i, 1] /
                                              combined[:, 1].sum() * 100))
                                + '%')
        df_dict = {item_name: item_list, 'Percentage': percent_list}
        df = pd.DataFrame(df_dict)
        return df

    def rdf_integral(self, species, distances, run_start, run_end):
        """ Calculate the integral of the radial distribution function of
        selected species

        Args:
            species (str): The interested species.
            distances (dict): A dict of coordination cutoff distance of species.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns pandas.dataframe of the species and the coordination number.
        """

        cn_values = self.coord_num_array_multi_species(species, distances,
                                                       run_start, run_end)
        item_name = "species in first solvation shell"
        item_list = []
        cn_list = []
        for kw in cn_values.keys():
            if kw != "total":
                shell_component, shell_count \
                    = np.unique(cn_values[kw].flatten(), return_counts=True)
                cn = (shell_component * shell_count / shell_count.sum()).sum()
                item_list.append(kw)
                cn_list.append(cn)
        df_dict = {item_name: item_list, self.name: cn_list}
        df = pd.DataFrame(df_dict)
        return df

    def shell_simple(self, species, distance, run_start, run_end):
        """ Tabulates the percentage of each solvation structures (CIP/SSIP/AGG)

        Args:
            species (str): The interested species.
            distance (int or float): The coordination cutoff distance.
            run_start (int): Start time step.
            run_end (int): End time step.

        Returns pandas.dataframe of the solvation structure and percentage.
        """
        num_array = self.coord_num_array_simple(species, distance,
                                                run_start, run_end)

        shell_component, shell_count = np.unique(num_array.flatten(),
                                                 return_counts=True)
        combined = np.vstack((shell_component, shell_count)).T

        item_name = "solvation structure"
        item_dict = {"1": "ssip", "2": "cip", "3": "agg"}
        item_list = []
        percent_list = []
        for i in range(len(combined)):
            item = str(int(combined[i, 0]))
            item_list.append(item_dict[item])
            percent_list.append(str("%.4f" % (combined[i, 1] /
                                              combined[:, 1].sum() * 100))
                                + '%')
        df_dict = {item_name: item_list, 'Percentage': percent_list}
        df = pd.DataFrame(df_dict)
        return df

    def get_msd_all(self, start=None, stop=None, fft=True):
        """ Calculates the mean square displacement (MSD) of the cation

        Args:
            start (int): Start time step.
            stop (int): End time step.
            fft (bool): Whether to use fft to calculate msd. Default to True.

        Returns an array of MSD values for each time in the trajectory
        """
        msd_array = total_msd(self.unwrapped_run, start=start, stop=stop,
                              select=self.select_dict["cation"], fft=fft)
        return msd_array

    def get_msd_by_length(self, distance, run_start, run_end):
        nvt_run = self.unwrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        free_array, attach_array = special_msd(nvt_run, li_atoms,
                                               self.select_dict, distance,
                                               run_start, run_end)
        return free_array, attach_array

    def get_msd_partial(self, distance, run_start, run_end, largest=1000):
        nvt_run = self.unwrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        free_array, attach_array = partial_msd(nvt_run, li_atoms, largest,
                                               self.select_dict, distance,
                                               run_start, run_end)
        return free_array, attach_array

    def get_d(self, msd_array, start, stop, percentage=1):
        """ Calculates the self-diffusion coefficient of the cation and
        the Nernst-Einstein conductivity

        Args:
            msd_array (numpy.array): msd array.
            start (int): Start time step.
            stop (int): End time step.
            percentage (int or float): The percentage of the cation.
                Default to 1.

        Print self-diffusion coefficient in m^2/s and NE conductivity in mS/cm.
        """
        a2 = 1e-20
        ps = 1e-12
        s_m_to_ms_cm = 10
        if percentage != 1:
            d = (msd_array[start] - msd_array[stop]) \
                / (start-stop) / self.time_step / 6 * a2 / ps
            sigma = percentage * d * self.d_to_sigma * s_m_to_ms_cm
            print("Diffusivity of", "%.2f" % (percentage * 100) + "% Li: ",
                  d, "m^2/s")
            print("NE Conductivity of", "%.2f" % (percentage * 100) + "% Li: ",
                  sigma, "mS/cm")
        else:
            d = (msd_array[start] - msd_array[stop]) \
                / (start - stop) / self.time_step / 6 * a2 / ps
            sigma = d * self.d_to_sigma * s_m_to_ms_cm
            print("Diffusivity of all Li:", d, "m^2/s")
            print("NE Conductivity of all Li:", sigma, "mS/cm")

    def get_neighbor_corr(self, species_list, distance, run_start, run_end):
        """ Calculates the neighbor auto-corelation function (ACF)
        of selected species around cation

        Args:
            species_list (list): List of species name.
            distance (int or float): Cutoff distance of neighbor.
            run_start: Start time step.
            run_end (int): End time step.

        Returns an array of the time series and a dict of ACFs of each species.
        """
        return calc_neigh_corr(self.wrapped_run, species_list, self.select_dict,
                               distance, self.time_step, run_start, run_end)

    @staticmethod
    def get_residence_time(species_list, times, acf_avg_dict, cutoff_time):
        """ Calculates the residence time of selected species around cation

        Args:
            species_list (list): List of species name.
            times (np.ndarray): The time series.
            acf_avg_dict: A dict of ACFs of each species.
            cutoff_time (int): Cutoff time for fitting the exponential decay.

        Returns the residence time of each species.
        """
        return fit_residence_time(times, species_list,
                                  acf_avg_dict, cutoff_time)

    def get_neighbor_trj(self, run_start, run_end, li_atom, species, distance):
        """ Calculates the distance of cation-neighbor as a function of time

        Args:
            run_start (int): start time step.
            run_end (int): end time step.
            li_atom (MDAnalysis.core.groups.Atom): the interested cation
                atom object.
            species (str): The interested neighbor species.
            distance (int or float): The coordination cutoff distance.

        Returns a dict of distance arrays of cation-neighbor
        as a function of time with neighbor id as keys.
        """
        return trajectory(self.wrapped_run, li_atom, run_start, run_end,
                          species, self.select_dict, distance)

    def get_hopping_freq_dist(self, run_start, run_end, species, distance,
                              hopping_cutoff, smooth=51):
        """ Calculates the cation hopping rate and hopping distance

        Args:
            run_start (int): Start time step.
            run_end (int): End time step.
            species (str): cation Binding site species.
            distance (int or float): Binding cutoff distance.
            hopping_cutoff: (int or float): Detaching cutoff distance.
            smooth (int): The length of the smooth filter window. Default to 51.

        Returns the cation average hopping rate and average hopping distance.
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        freqs = []
        hopping_distance = []
        for li in tqdm_notebook(li_atoms[:]):
            neighbor_trj = trajectory(nvt_run, li, run_start, run_end, species,
                                      self.select_dict, distance)
            sites, freq, steps = find_nearest(neighbor_trj, self.time_step,
                                              distance, hopping_cutoff,
                                              smooth=smooth)

            coords = []
            for step in steps:
                coord_li = nvt_run.trajectory[step + 1000][li.id - 1]
                coords.append(coord_li)
            if len(coords) > 1:
                dists = []
                for i in range(len(coords) - 1):
                    dist = distance_array(coords[i + 1], coords[i],
                                          box=self.get_nvt_dimension())[0][0]
                    dists.append(dist)
                li_mean_dists = np.mean(dists)
                hopping_distance.append(li_mean_dists)
            freqs.append(freq)
        return np.mean(freqs), np.mean(hopping_distance)

    def get_heat_map(self, run_start, run_end, species, distance,
                     hopping_cutoff, cartesian_by_ref=None, bind_atom_type=None,
                     sym_dict=None, sample=None, smooth=51):
        """ Calculates the heatmap matrix of cation around a cluster/molecule

        Args:
            run_start (int): Start time step.
            run_end (int): End time step.
            species (str): cation Binding site species.
            distance (int or float): Binding cutoff distance.
            hopping_cutoff: (int or float): Detaching cutoff distance.
            cartesian_by_ref (np.array): Transformation matrix between cartesian
                and reference coordinate systems. Default to None.
            bind_atom_type (str): Selection for binding site atom.
                Default to None.
            sym_dict (dict): Dictionary of symmetry operation dictionary.
                Default to None.
            sample (int): Number of samples desired. Default to None,
                which is no sampling.
            smooth (int): The length of the smooth filter window. Default to 51.
        """
        nvt_run = self.wrapped_run
        li_atoms = nvt_run.select_atoms(self.select_dict["cation"])
        coord_list = np.array([[0, 0, 0]])
        for li in tqdm_notebook(li_atoms[:]):
            neighbor_trj = trajectory(nvt_run, li, run_start, run_end, species,
                                      self.select_dict, distance)
            sites, freq, steps = find_nearest(neighbor_trj, self.time_step,
                                              distance, hopping_cutoff,
                                              smooth=smooth)
            if bind_atom_type is None:
                bind_atom_type = self.select_dict["anion"]
            if cartesian_by_ref is None:
                cartesian_by_ref = np.array([[1, 0, 0],
                                             [0, 1, 0],
                                             [0, 0, 1]])
            coords = heat_map(nvt_run, li, sites, 4, bind_atom_type,
                              cartesian_by_ref, run_start, run_end)
            coord_list = np.concatenate((coord_list, coords), axis=0)
        if sym_dict:
            return get_full_coords(coord_list, **sym_dict, sample=sample)
        else:
            return get_full_coords(coord_list, sample=sample)

    def get_cluster_distance(self, run_start, run_end, neighbor_cutoff,
                             cluster_center="center"):
        """ Calculates the average distance of the center of clusters/molecules

        Args:
            run_start (int): Start time step.
            run_end (int): End time step.
            neighbor_cutoff (int of float): Upper limit of
                first nearest neighbor.
            cluster_center (str): species name of cluster center.
                Default to "center".
        """
        center_atoms = \
            self.wrapped_run.select_atoms(self.select_dict[cluster_center])
        trj = self.wrapped_run.trajectory[run_start:run_end:]
        means = []
        for ts in trj:
            distance_matrix = capped_distance(center_atoms.positions,
                                              center_atoms.positions,
                                              max_cutoff=neighbor_cutoff,
                                              box=ts.dimensions,
                                              return_distances=True)[1]
            distance_matrix[distance_matrix == 0] = np.nan
            means.append(np.nanmean(distance_matrix))
        return np.mean(means)



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

"""
This module implements a core class PackmolWrapper for packing molecules
into a single box.

You need the Packmol package to run the code, see
http://m3g.iqm.unicamp.br/packmol or
http://leandro.iqm.unicamp.br/m3g/packmol/home.shtml
for download and setup instructions. You may need to manually
set the folder of the packmol executable to the PATH environment variable.
"""

import subprocess
import os

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


class PackmolWrapper:
    """
    Wrapper for the Packmol software that can be used to pack various types of
    molecules into a one single unit.

    Args:
        path (str): The path to the directory for file i/o.
        structures (list): A list containing dict information of molecules.
            Each molecule have two keys, "name", the structure name,
            and "file", the path to the structure file.
        numbers (dict): A dict of the numbers of each molecule
        box (list): A list of xlo, ylo, zlo, xhi, yhi, zhi, in Å.
        tolerance (float): Tolerance for packmol.
        seed (int): Random seed for packmol.
        inputfile (str): Path to the input file. Default to 'packmol.inp'.
        outputfile (str): Path to the output file. Default to 'output.xyz'.

    Examples:

        >>> structures = [{"name": "structure_name", "file": "/path/to/xyz"}]
        >>> pw = PackmolWrapper("/path/to/work/dir",
        ...                     structures,
        ...                     {"structure_name": '2'},
        ...                     [0., 0., 0., 10., 10., 10.]
        ... )
        >>> pw.make_packmol_input()
        >>> pw.run_packmol()
    """
    def __init__(self, path, structures, numbers, box, tolerance=None,
                 seed=None, inputfile='packmol.inp', outputfile='output.xyz'):
        self.path = path
        self.input = os.path.join(self.path, inputfile)
        self.output = os.path.join(self.path, outputfile)
        self.screen = os.path.join(self.path, 'packmol.stdout')
        self.structures = structures
        self.numbers = numbers
        self.box = box
        if tolerance is None:
            self.tolerance = 2.0
        else:
            self.tolerance = tolerance
        if seed is None:
            self.seed = 123
        else:
            self.seed = seed

    def run_packmol(self):
        """Run packmol and write out the packed structure."""
        try:
            p = subprocess.run('packmol < {}'.format(self.input),
                               check=True,
                               shell=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as e:
            raise ValueError("Packmol failed with errorcode {}"
                             " and stderr: {}".format(e.returncode, e.stderr))
        else:
            with open(self.screen, 'w') as out:
                out.write(p.stdout.decode())

    def make_packmol_input(self):
        """Make a Packmol usable input file."""

        with open(self.input, 'w') as out:
            out.write("# " + ' + '.join(self.numbers[structure["name"]] + " "
                                        + structure["name"] for structure
                                        in self.structures) + "\n")
            out.write("# Packmol input generated by mdgo.\n")
            out.write('seed {}\n'.format(self.seed))
            out.write('tolerance {}\n\n'.format(self.tolerance))
            out.write('filetype xyz\n\n')
            out.write('output {}\n\n'.format(self.output))

            for structure in self.structures:
                out.write("structure {}\n".format(structure["file"]))
                out.write("  number {}\n".format(
                    self.numbers[structure["name"]]))
                out.write("  inside box {}\n".format(
                    " ".join(str(i) for i in self.box)))
                out.write("end structure\n\n")


if __name__ == "__main__":
    structures = [{"name": "EMC",
                   "file": "/Users/th/Downloads/test_selenium/EMC.lmp.xyz"}]
    pw = PackmolWrapper("/Users/th/Downloads/test_selenium/", structures,
                        {"EMC": '2'}, [0., 0., 0., 10., 10., 10.])
    pw.make_packmol_input()
    pw.run_packmol()



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import numpy as np
from tqdm import tqdm_notebook
from scipy import stats

__author__ = "Kara Fong, Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


def autocorrFFT(x):
    """
    Calculates the position autocorrelation function using
    the fast Fourier transform.
    """
    N = len(x)
    F = np.fft.fft(x, n=2 * N)  # 2*N because of zero-padding
    PSD = F * F.conjugate()
    res = np.fft.ifft(PSD)
    res = (res[:N]).real  # now we have the autocorrelation in convention B
    n = N * np.ones(N) - np.arange(0, N)  # divide res(m) by (N-m)
    return res / n  # this is the autocorrelation in convention A


def msd_fft(r):
    """
    Calculates mean square displacement of the array r using
    the fast Fourier transform.
    """
    N = len(r)
    D = np.square(r).sum(axis=1)
    D = np.append(D, 0)
    S2 = sum([autocorrFFT(r[:, i]) for i in range(r.shape[1])])
    Q = 2 * D.sum()
    S1 = np.zeros(N)
    for m in range(N):
        Q = Q - D[m - 1] - D[N - m]
        S1[m] = Q / (N - m)
    return S1 - 2 * S2


def calc_cond(u, anions, cations, run_start, cation_charge=1, anion_charge=-1):
    """ Calculates the conductivity "mean square displacement" given
    an MDAnalysis universe (u) and a selection of atoms or molecules (sel)

    Args:
    u: MDAnalysis Universe
    sel: Selection of atoms or molecules. Should be
    a MDAnalysis AtomGroup.

    Returns an array of MSD values for each time in the trajectory.

    Note:
        Coordinates must be unwrapped (in dcd file when creating MDAnalysis
        Universe)
    """
    # Current code assumes anion and cation selections are single atoms
    qr = []
    for ts in tqdm_notebook(u.trajectory[run_start:]):
        qr_temp = np.zeros(3)
        for anion in anions.atoms:
            qr_temp += anion.position * anion_charge
        for cation in cations.atoms:
            qr_temp += cation.position * cation_charge
        qr.append(qr_temp)
    msd = (msd_fft(np.array(qr)))
    return msd


def conductivity_calculator(time_array, cond_array, v, name, start, end):
    # Unit conversions
    A2cm = 1e-8
    ps2s = 1e-12
    e2c = 1.60217662e-19
    convert = e2c*e2c/ps2s/A2cm*1000  # mS/cm
    kb = 1.38064852e-23
    T = 298.15

    # Calculate conductivity of mof
    slope_cond_avg, intercept_cond_avg, r_value, p_value, std_err \
        = stats.linregress(time_array[start:end], cond_array[start:end])
    cond_einstein_mof = slope_cond_avg/6/kb/T/v*convert
    error_mof = std_err/6/kb/T/v*convert

    print("GK Conductivity of " + name + ": " + str(cond_einstein_mof)
          + " ± " + str(error_mof) + " mS/cm")



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import re
import numpy as np
import matplotlib.pyplot as plt
from statsmodels.tsa.stattools import acovf
from scipy.optimize import curve_fit
from tqdm import tqdm_notebook

__author__ = "Kara Fong, Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"


# Create adjacency matrix for one li
def neighbors_one_li(nvt_run, li_atom, species, selection_dict, distance,
                     run_start, run_end):
    bool_values = dict()
    time_count = 0
    for ts in nvt_run.trajectory[run_start:run_end:]:
        if species in selection_dict.keys():
            selection = "(" + selection_dict[species] + ") and (around " \
                        + str(distance) + " index " \
                        + str(li_atom.id - 1) + ")"
            shell = nvt_run.select_atoms(selection)
        else:
            print('Invalid species selection')
            return None
        for atom in shell.atoms:
            if str(atom.id) not in bool_values:
                bool_values[str(atom.id)] = np.zeros(int((run_end-run_start)/1))
            bool_values[str(atom.id)][time_count] = 1
        time_count += 1
    return bool_values


# Calculate ACF
def calc_acf(a_values):
    acfs = []
    for atom_id, neighbors in a_values.items():
        atom_id = int(re.search(r'\d+', atom_id).group())
        acfs.append(acovf(neighbors, demean=False, unbiased=True, fft=True))
    return acfs


def exponential_func(x, a, b, c):
    return a * np.exp(-b * x) + c


def calc_neigh_corr(nvt_run, species_list, selection_dict, distance, time_step,
                    run_start, run_end):
    # Set up times array
    times = []
    step = 0
    li_atoms = nvt_run.select_atoms(selection_dict["cation"])
    for ts in nvt_run.trajectory[run_start:run_end]:
        times.append(step * time_step)
        step += 1
    times = np.array(times)

    acf_avg = dict()
    for kw in species_list:
        acf_all = list()
        for li in tqdm_notebook(li_atoms[::]):
            adjacency_matrix = neighbors_one_li(nvt_run, li, kw, selection_dict,
                                                distance, run_start, run_end)
            acfs = calc_acf(adjacency_matrix)
            [acf_all.append(acf) for acf in acfs]
        acf_avg[kw] = np.mean(acf_all, axis=0)
    return times, acf_avg


def fit_residence_time(times, species_list, acf_avg_dict, cutoff_time):
    acf_avg_norm = dict()
    popt = dict()
    pcov = dict()
    tau = dict()

    # Exponential fit of solvent-Li ACF
    for kw in species_list:
        acf_avg_norm[kw] = acf_avg_dict[kw] / acf_avg_dict[kw][0]
        popt[kw], pcov[kw] = curve_fit(exponential_func, times[:cutoff_time],
                                       acf_avg_norm[kw][:cutoff_time],
                                       p0=(1, 1e-4, 0))
        tau[kw] = 1 / popt[kw][1]  # ps

    # Plot ACFs
    colors = ["b", "g", "r", "c", "m", "y"]
    line_styles = ['-', '--', '-.', ':']
    for i, kw in enumerate(species_list):
        plt.plot(times, acf_avg_norm[kw], label=kw, color=colors[i])
        plt.plot(np.linspace(0, 10000, 1000),
                 exponential_func(np.linspace(0, 10000, 1000), *popt[kw]),
                 line_styles[i], color='k', label=kw + ' Fit')

    plt.xlabel('Time (ps)')
    plt.legend()
    plt.ylabel('Neighbor Autocorrelation Function')
    plt.ylim(0, 1)
    plt.xlim(0, 10000)
    plt.show()

    return tau



# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

import os

from setuptools import setup, find_packages

module_dir = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(module_dir, "README.md"), 'r') as f:
    readme = f.read()

INSTALL_REQUIRES = [
    "numpy>=1.16.0",
    "pandas",
    "matplotlib",
    "scipy",
    "tqdm",
    "pymatgen",
    "statsmodels",
    "pubchempy",
    "MDAnalysis",
    "selenium"
]

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    INSTALL_REQUIRES = []

if __name__ == "__main__":
    setup(
        name='mdgo',
        version='0.1.0',
        description='A codebase for MD simulation setup and results analysis.',
        long_description=readme,
        long_description_content_type="text/markdown",
        license="MIT",
        author="mdgo development team",
        author_email='tingzheng_hou@berkeley.edu',
        maintainer="Tingzheng Hou",
        maintainer_email='tingzheng_hou@berkeley.edu',
        url="https://github.com/htz1992213/mdgo",
        packages=find_packages(),
        install_requires=INSTALL_REQUIRES,
        extras_require={
            'web': [
                'sphinx',
                'sphinx_rtd_theme',
            ]
        },
        python_requires='>=3.6'
    )




# coding: utf-8
# Copyright (c) Tingzheng Hou.
# Distributed under the terms of the MIT License.

"""
This module implements two core class FFcrawler and MaestroRunner
for generating LAMMPS/GROMACS data files from molecule structure using
the LigParGen web server and Maestro, respectively.

For using the FFcrawler class:

  * Download the ChromeDriver executable that
    matches your Chrome version via https://chromedriver.chromium.org/downloads

For using the MaestroRunner class:

  * Download a free Maestro via https://www.schrodinger.com/freemaestro

  * Install the package and set the environment variable $SCHRODINGER
    (e.g. 'export SCHRODINGER=/opt/schrodinger/suites2020-4', please
    check https://www.schrodinger.com/kb/446296 or
    https://www.schrodinger.com/kb/1842 for details.

"""

from pymatgen.io.lammps.data import LammpsData
from mdgo.util import mass_to_name, ff_parser
import pubchempy as pcp
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from string import Template
from urllib.parse import quote
import time
import os
import shutil
import signal
import subprocess

__author__ = "Tingzheng Hou"
__version__ = "1.0"
__maintainer__ = "Tingzheng Hou"
__email__ = "tingzheng_hou@berkeley.edu"
__date__ = "Feb 9, 2021"

MAESTRO = "$SCHRODINGER/maestro -console -nosplash"
FFLD = "$SCHRODINGER/utilities/ffld_server -imae {} " \
       "-version 14 -print_parameters -out_file {}"


class FFcrawler:
    """
    Web scrapper that can automatically upload structure to the LigParGen
    server and download LAMMPS/GROMACS data file.

    Args:
        write_dir (str): Directory for writing output.
        chromedriver_dir (str): Directory to the ChromeDriver executable.
        headless (bool): Whether to run Chrome in headless (silent) mode.
            Default to True.
        xyz (bool): Whether to write the structure in the LigParGen
            generated data file as .xyz. Default to False. This is useful
            because the order and the name of the atoms could be
            different from the initial input.)
        gromacs (bool): Whether to save GROMACS format data files.
            Default to False.

    Examples:

        >>> lpg = FFcrawler('/path/to/work/dir', '/path/to/chromedriver')
        >>> lpg.data_from_pdb("/path/to/pdb")
    """

    def __init__(
            self,
            write_dir,
            chromedriver_dir,
            headless=True,
            xyz=False,
            gromacs=False
    ):
        """Base constructor."""
        self.write_dir = write_dir
        self.xyz = xyz
        self.gromacs = gromacs
        self.preferences = {"download.default_directory": write_dir,
                            "safebrowsing.enabled": "false",
                            "profile.managed_default_content_settings.images":
                                2}
        self.options = webdriver.ChromeOptions()
        self.options.add_argument('user-agent="Mozilla/5.0 '
                                  '(Macintosh; Intel Mac OS X 10_14_6) '
                                  'AppleWebKit/537.36 (KHTML, like Gecko) '
                                  'Chrome/88.0.4324.146 Safari/537.36"')
        self.options.add_argument("--window-size=1920,1080")
        if headless:
            self.options.add_argument('--headless')
        self.options.add_experimental_option("prefs", self.preferences)
        self.options.add_experimental_option('excludeSwitches',
                                             ['enable-automation'])
        self.web = webdriver.Chrome(chromedriver_dir, options=self.options)
        self.wait = WebDriverWait(self.web, 10)
        self.web.get("http://zarbi.chem.yale.edu/ligpargen/")
        time.sleep(1)
        print("LigParGen server connected.")

    def data_from_pdb(self, pdb_dir):
        """
        Use the LigParGen server to generate a LAMMPS data file from a pdb file.

        Arg:
            pdb_dir (str): The path to the input pdb structure file.

        Write out a LAMMPS data file.
        """
        upload = self.web.find_element_by_xpath('//*[@id="exampleMOLFile"]')
        upload.send_keys(pdb_dir)
        submit = self.web.find_element_by_xpath(
            '/html/body/div[2]/div/div[2]/form/button[1]')
        submit.click()
        pdb_filename = os.path.basename(pdb_dir)
        try:
            self.download_data(os.path.splitext(pdb_filename)[0] + ".lmp")
        except TimeoutException:
            print(
                "Timeout! Web server no response for 10s, file download failed!"
            )
        finally:
            self.web.quit()

    def data_from_smiles(self, smiles_code):
        """
        Use the LigParGen server to generate a LAMMPS data file
        from a SMILES code.

        Arg:
            smiles_code (str): The SMILES code for the LigParGen input.

        Write out a LAMMPS data file.
        """
        smile = self.web.find_element_by_xpath('//*[@id="smiles"]')
        smile.send_keys(smiles_code)
        submit = self.web.find_element_by_xpath(
            '/html/body/div[2]/div/div[2]/form/button[1]')
        submit.click()
        try:
            self.download_data(smiles_code + '.lmp')
        except TimeoutException:
            print(
                "Timeout! Web server no response for 10s, file download failed!"
            )
        finally:
            self.web.quit()

    def download_data(self, lmp_name):
        """
        Helper function that download and write out the LAMMPS data file.

        Arg:
            lmp_name (str): Name of the LAMMPS data file.
        """
        print("Structure info uploaded. Rendering force field...")
        self.wait.until(
            EC.presence_of_element_located((By.NAME, 'go'))
        )
        data_lmp = self.web.find_element_by_xpath(
            "/html/body/div[2]/div[2]/div[1]/div/div[14]/form/input[1]"
        )
        data_lmp.click()
        print("Force field file downloaded.")
        time.sleep(1)
        lmp_file = max(
            [self.write_dir + "/" + f for f
             in os.listdir(self.write_dir)
             if os.path.splitext(f)[1] == ".lmp"],
            key=os.path.getctime)
        if self.xyz:
            data = LammpsData.from_file(lmp_file)
            element_id_dict = mass_to_name(data.masses)
            coords = data.atoms[['type', 'x', 'y', 'z']]
            lines = list()
            lines.append(str(len(coords.index)))
            lines.append("")
            for _, r in coords.iterrows():
                line = element_id_dict.get(int(r['type'])) + ' ' + ' '.join(
                    str(r[loc]) for loc in ["x", "y", "z"])
                lines.append(line)

            with open(os.path.join(self.write_dir, lmp_name + ".xyz"),
                      "w") as xyz_file:
                xyz_file.write("\n".join(lines))
            print(".xyz file saved.")
        if self.gromacs:
            data_gro = self.web.find_element_by_xpath(
                "/html/body/div[2]/div[2]/div[1]/div/div[8]/form/input[1]"
            )
            data_itp = self.web.find_element_by_xpath(
                "/html/body/div[2]/div[2]/div[1]/div/div[9]/form/input[1]"
            )
            data_gro.click()
            data_itp.click()
            time.sleep(1)
            gro_file = max(
                [self.write_dir + "/" + f for f
                 in os.listdir(self.write_dir)
                 if os.path.splitext(f)[1] == ".gro"],
                key=os.path.getctime)
            itp_file = max(
                [self.write_dir + "/" + f for f
                 in os.listdir(self.write_dir)
                 if os.path.splitext(f)[1] == ".itp"],
                key=os.path.getctime)
            shutil.move(
                gro_file,
                os.path.join(self.write_dir, lmp_name[:-4] + ".gro")
            )
            shutil.move(
                itp_file,
                os.path.join(self.write_dir, lmp_name[:-4] + ".itp")
            )
        shutil.move(lmp_file, os.path.join(self.write_dir, lmp_name))
        print("Force field file saved.")


class MaestroRunner:

    """
    Wrapper for the Maestro software that can be used to generate the OPLS_2005
    force field parameter for a molecule.

    Args:
        structure_dir (str): Path to the structure file.
            Supported input format please check
            https://www.schrodinger.com/kb/1278
        working_dir (str): Directory for writing intermediate
            and final output.
        out (str): Force field output form. Default to "lmp",
            the data file for LAMMPS. Other supported formats
            are under development.
        cmd_template (str): String template for input script
            with placeholders. Default to None, i.e., using
            the default template.
        assign_bond (bool): Whether to assign bond to the input
            structure. Default to None.

    Supported input format please check https://www.schrodinger.com/kb/1278

    The OPLS_2005 parameters are described in

    Banks, J.L.; Beard, H.S.; Cao, Y.; Cho, A.E.; Damm, W.; Farid, R.;
    Felts, A.K.; Halgren, T.A.; Mainz, D.T.; Maple, J.R.; Murphy, R.;
    Philipp, D.M.; Repasky, M.P.; Zhang, L.Y.; Berne, B.J.; Friesner, R.A.;
    Gallicchio, E.; Levy. R.M. Integrated Modeling Program, Applied Chemical
    Theory (IMPACT). J. Comp. Chem. 2005, 26, 1752.

    The OPLS_2005 parameters are located in

    $SCHRODINGER/mmshare-vversion/data/f14/

    Examples:

        >>> mr = MaestroRunner('/path/to/structure', '/path/to/working/dir')
        >>> mr.get_mae()
        >>> mr.get_ff()
    """

    template_assignbond = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "templates",
        "mae_cmd_assignbond.txt"
    )

    template_noassignbond = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "templates",
        "mae_cmd_noassignbond.txt"
    )

    def __init__(
            self,
            structure_dir,
            working_dir,
            out="lmp",
            cmd_template=None,
            assign_bond=False
    ):
        """Base constructor."""
        self.structure = structure_dir
        self.out = out
        self.structure_format = os.path.splitext(self.structure)[1][1:]
        self.name = os.path.splitext(os.path.split(self.structure)[-1])[0]
        print("Input format:", self.structure_format)
        self.work = working_dir
        self.cmd = os.path.join(self.work, "maetro_script.cmd")
        self.mae = os.path.join(self.work, self.name)
        self.ff = os.path.join(self.work, self.name + ".out")
        self.xyz = os.path.join(self.work, self.name + ".xyz")
        if cmd_template:
            self.cmd_template = cmd_template
        else:
            if assign_bond:
                with open(self.template_assignbond, "r") as f:
                    cmd_template = f.read()
                self.cmd_template = cmd_template
            else:
                with open(self.template_noassignbond, "r") as f:
                    cmd_template = f.read()
                self.cmd_template = cmd_template

    def get_mae(self):
        """Write a Maestro command script and execute it to generate a
        maestro file containing all the info needed."""
        with open(self.cmd, 'w') as f:
            cmd_template = Template(self.cmd_template)
            cmd_script = cmd_template.substitute(
                file=self.structure,
                mae=self.mae,
                xyz=self.xyz
            )
            f.write(cmd_script)
        try:
            p = subprocess.Popen(
                f'{MAESTRO} -c {self.cmd}',
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                preexec_fn=os.setsid
            )

            counter = 0
            while not os.path.isfile(self.mae + ".mae"):
                time.sleep(1)
                counter += 1
                if counter > 30:
                    raise TimeoutError(
                        "Failed to generate Maestro file in 30 secs!"
                    )
            print("Maestro file generated.")

        except subprocess.CalledProcessError as e:
            raise ValueError(
                "Maestro failed with errorcode {}  and stderr: {}".format(
                    e.returncode, e.stderr
                )
            )
        finally:
            os.killpg(os.getpgid(p.pid), signal.SIGTERM)

    def get_ff(self):
        """Read the Maestro file and save the force field as LAMMPS data file.
        """
        try:
            p = subprocess.run(
                FFLD.format(self.mae + ".mae", self.ff),
                check=True,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
        except subprocess.CalledProcessError as e:
            raise ValueError(
                "Maestro failed with errorcode {} and stderr: {}".format(
                    e.returncode, e.stderr
                )
            )
        print("Maestro force field file generated.")
        if self.out:
            if self.out == "lmp":
                with open(
                        os.path.join(self.work, self.name + "." + self.out), 'w'
                ) as f:
                    f.write(ff_parser(self.ff, self.xyz))
                print("LAMMPS data file generated.")
            else:
                print("Output format not supported, ff format not converted.")


class PubChemRunner:

    """Wrapper for accessing PubChem data that can be used to retriving compound
    structure and information.

    Args:
        write_dir (str): Directory for writing output.
        chromedriver_dir (str): Directory to the ChromeDriver executable.
        api (bool): Whether to use the PUG REST web interface for accessing
            PubChem data. If None, then all search/download will be
            performed via web browser mode. Default to True.
        headless (bool): Whether to run Chrome in headless (silent) mode.
            Default to False.

    Examples:
        >>> web = PubChemRunner('/path/to/work/dir', '/path/to/chromedriver')
        >>> long_name, short_name = "ethylene carbonate", "PC"
        >>> cid = web.obtain_entry(long_name, short_name)
    """

    def __init__(
            self,
            write_dir,
            chromedriver_dir,
            api=True,
            headless=False,
    ):
        """Base constructor."""
        self.write_dir = write_dir
        self.api = api
        if not self.api:
            self.preferences = {
                "download.default_directory": write_dir,
                "safebrowsing.enabled": "false",
                "profile.managed_default_content_settings.images": 2
            }
            self.options = webdriver.ChromeOptions()
            self.options.add_argument(
                'user-agent="Mozilla/5.0 '
                '(Macintosh; Intel Mac OS X 10_14_6) '
                'AppleWebKit/537.36 (KHTML, like Gecko) '
                'Chrome/88.0.4324.146 Safari/537.36"'
            )
            self.options.add_argument("--window-size=1920,1080")
            if headless:
                self.options.add_argument('--headless')
            self.options.add_experimental_option("prefs", self.preferences)
            self.options.add_experimental_option('excludeSwitches',
                                                 ['enable-automation'])
            self.web = webdriver.Chrome(chromedriver_dir, options=self.options)
            self.wait = WebDriverWait(self.web, 10)
            self.web.get("https://pubchem.ncbi.nlm.nih.gov/")
            time.sleep(1)
            print("PubChem server connected.")

    def obtain_entry(self, search_text, name, output_format='sdf'):
        """
        Search the PubChem database with a text entry and save the
        structure in desired format.

        Args:
            search_text (str): The text to use as a search query.
            name (str): The short name for the molecule.
            output_format (str): The output format of the structure.
                Default to sdf.
        """
        if self.api:
            return self._obtain_entry_api(
                search_text,
                name,
                output_format=output_format
            )
        else:
            return self._obtain_entry_web(
                search_text,
                name,
                output_format=output_format
            )

    def _obtain_entry_web(self, search_text, name, output_format):
        cid = None
        try:
            query = quote(search_text)
            url = "https://pubchem.ncbi.nlm.nih.gov/#query=" + query
            self.web.get(url)
            time.sleep(1)
            best_xpath = (
                '//*[@id="featured-results"]/div/div[2]'
                '/div/div[1]/div[2]/div[1]/a/span/span'
            )
            relevant_xpath = (
                '//*[@id="collection-results-container"]'
                '/div/div/div[2]/ul/li[1]/div/div/div[1]'
                '/div[2]/div[1]/a/span/span'
            )
            if EC.presence_of_element_located((By.XPATH, best_xpath)):
                match = self.web.find_element_by_xpath(best_xpath)
            else:
                match = self.web.find_element_by_xpath(relevant_xpath)
            match.click()
            cid_locator = (
                '//*[@id="main-content"]/div/div/div[1]/'
                'div[3]/div/table/tbody/tr[1]/td'
            )
            self.wait.until(
                EC.presence_of_element_located((By.XPATH, cid_locator))
            )
            cid = self.web.find_element_by_xpath(cid_locator).text
            print("Best match found, PubChem ID:", cid)
            self.web.get(
                f'https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/CID/{cid}/'
                f'record/{output_format.upper()}/?record_type=3d&'
                f'response_type=save&response_basename={name + "_" + cid}'
             )
            print("Waiting for downloads.", end="")
            time.sleep(1)
            while any([filename.endswith(".crdownload") for filename in
                       os.listdir(self.write_dir)]):
                time.sleep(1)
                print(".", end="")
            print("\nStructure file saved.")
        except TimeoutException:
            print(
                "Timeout! Web server no response for 10s, "
                "file download failed!"
            )
        except NoSuchElementException:
            print(
                "The download link was not correctly generated, "
                "file download failed!\n"
                "Please try another search text or output format."
            )
        finally:
            self.web.quit()
        return cid

    def _obtain_entry_api(self, search_text, name, output_format):
        cid = None
        cids = pcp.get_cids(search_text, 'name', record_type='3d')
        if len(cids) == 0:
            print("No exact match found, please try the web search")
        else:
            cid = str(cids[0])
            pcp.download(
                output_format.upper(),
                os.path.join(
                    self.write_dir,
                    name + '_' + cid + '.' + output_format.lower()
                ),
                cid,
                record_type='3d',
                overwrite=True
            )
        return cid


if __name__ == "__main__":
    MR = MaestroRunner("/Users/th/Downloads/test_mr/EC_7303.sdf",
                       "/Users/th/Downloads/test_mr")
    MR.get_mae()
    MR.get_ff()
    """
    pcr = PubChemRunner(
        "/Users/th/Downloads/test_pc/",
        "/Users/th/Downloads/package/chromedriver/chromedriver",
        api=True
    )
    long_name = "ethylene carbonate"
    short_name = "EC"
    cid = pcr.obtain_entry(long_name, short_name, "sdf")

    
    LPG = FFcrawler(
        "/Users/th/Downloads/test_selenium",
        "/Users/th/Downloads/package/chromedriver/chromedriver",
        xyz=True,
        gromacs=True
    )
    LPG.data_from_pdb("/Users/th/Downloads/test_selenium/EMC.pdb")

    MR = MaestroRunner("/Users/th/Downloads/test_mr/EC.sdf",
                       "/Users/th/Downloads/test_mr")
    MR.get_mae()
    MR.get_ff()

    pcr = PubChemCrawler(
        "/Users/th/Downloads/test_pc/",
        "/Users/th/Downloads/package/chromedriver/chromedriver",
        headless=True
    )
    long_name = "Propylene Carbonate"
    short_name = "PC"
    cid = pcr.obtain_entry(long_name, short_name)
    MR = MaestroRunner(
        f"/Users/th/Downloads/test_pc/{short_name}_{cid}.sdf",
        "/Users/th/Downloads/test_pc")
    MR.get_mae()
    MR.get_ff()
    """
